from qdep.internal.common import *
from qdep.internal.prf import *


version = "1.1.1"


def prfgen(script_path, qmake="qmake", data_dir=None):
	if data_dir is not None:
		prf_path = data_dir
	else:
		qmake_res = sub_run([qmake, "-query", "QT_HOST_DATA"], check=True, stdout=subprocess.PIPE, encoding="UTF-8")
		prf_path = str(qmake_res.stdout).strip()
	prf_path = path.join(prf_path, "mkspecs", "features", "qdep.prf")
	print("Generating PRF-File as: ", prf_path)
	os.makedirs(path.dirname(prf_path), exist_ok=True)
	with open(prf_path, "w") as prf_file:
		prf_file.write("# Generated by {}\n".format(script_path))
		prf_file.write("qdep_force|!isEmpty(OUT_PWD) {  # Used to detect if qdep was loaded from an indirect evaluation\n\n")
		prf_file.write("isEmpty(QDEP_PATH) {\n")
		prf_file.write("\tQDEP_PATH = {}\n".format(script_path))
		prf_file.write("\twin32: QDEP_PATH = $${QDEP_PATH}.exe\n")
		prf_file.write("}\n")
		prf_file.write("isEmpty(QDEP_VERSION): QDEP_VERSION = {}\n".format(version))
		prf_file.write(qdep_prf)
		prf_file.write("\n}\n")


def init(profile):
	with open(profile, "a") as pro_file:
		pro_file.write("\nQDEP_DEPENDS += \n\n")
		pro_file.write("!load(qdep):error(\"Failed to load qdep feature! Run 'qdep prfgen --qmake $$QMAKE_QMAKE' to create it.\")\n")


def lupdate(pri_path, qmake="qmake", lupdate_args=None):
	if lupdate_args is None:
		lupdate_args = []

	qmake_res = sub_run([qmake, "-query", "QT_HOST_BINS"], check=True, stdout=subprocess.PIPE, encoding="UTF-8")
	lupdate_path = path.join(str(qmake_res.stdout).strip(), "lupdate")
	with tempfile.TemporaryDirectory() as tmp_dir:
		tmp_pro_path = path.join(tmp_dir, "lupdate.pro")
		with open(tmp_pro_path, "w") as tmp_file:
			tmp_file.write("include({})\n".format(pri_path))
			tmp_file.write("TRANSLATIONS = $$QDEP_TRANSLATIONS\n")
		sub_run([lupdate_path] + lupdate_args + [tmp_pro_path], check=True)


def clear(no_confirm=False):
	def folder_size(path):
		total = 0
		for entry in os.scandir(path):
			if entry.is_file():
				total += entry.stat().st_size
			elif entry.is_dir():
				total += folder_size(entry.path)
		return total

	if not no_confirm:
		print("All caches sources will be removed and have to be downloaded again. Make sure no other qdep instance is currently running!")
		ok = input("Do you really want ro remove all cached sources? [y/N]").lower()
		if len(ok) == 0:
			return
		elif ok != "y" and ok != "yes":
			return

	cache_dir = os.getenv("QDEP_CACHE_DIR", get_cache_dir_default())
	cache_dir = path.join(cache_dir, "src")
	if path.exists(cache_dir):
		rm_size = folder_size(cache_dir)
		shutil.rmtree(cache_dir)
	else:
		rm_size = 0
	print("Removed {} bytes".format(rm_size))


def versions(package, tags=True, branches=False, short=False, limit=None):
	package_url, _v, _p = package_resolve(package)

	if tags:
		pkg_tags = get_all_tags(package_url, tags=True, branches=False, allow_empty=True)
		if limit is not None:
			pkg_tags = pkg_tags[-limit:]
	else:
		pkg_tags = []

	if branches:
		pkg_branches = get_all_tags(package_url, tags=False, branches=True, allow_empty=True)
		if limit is not None:
			pkg_branches = pkg_branches[-limit:]
	else:
		pkg_branches = []

	if short:
		print(" ".join(pkg_branches + pkg_tags))
	else:
		if branches:
			print("Branches:")
			if len(pkg_branches) > 0:
				for branch in pkg_branches:
					print("  " + branch)
			else:
				print(" -- No branches found --")

		if tags:
			if branches:
				print("")

			print("Tags:")
			if len(pkg_tags) > 0:
				for tag in pkg_tags:
					print("  " + tag)
			else:
				print(" -- No tags found --")


def query(package, check=True, print_versions=False, expand=False):
	pkg_url, pkg_version, pkg_path = package_resolve(package)
	if check:
		if pkg_version is None:
			pkg_version = get_latest_tag(pkg_url, allow_empty=True, allow_error=True)
			pkg_exists = pkg_version is not None
		else:
			pkg_exists = pkg_version in get_all_tags(pkg_url, branches=True, tags=True, allow_empty=True, allow_error=True)
	else:
		pkg_exists = False  # not knowing means false

	if expand:
		if pkg_version is None:
			raise Exception("Unable to determine package version")
		else:
			print("{}@{}{}".format(pkg_url, pkg_version, pkg_path))
	else:
		print("Input:", package)
		if pkg_version is not None:
			print("Expanded Name: {}@{}{}".format(pkg_url, pkg_version, pkg_path))
		else:
			print("Expanded Name:", None)
		print("URL:", pkg_url)
		print("Version:", pkg_version)
		print("Path:", pkg_path)
		if check:
			print("Exists:", pkg_exists)

		if print_versions:
			print("")
			versions(package, branches=True)


def get(*targets, extract=False, evaluate=False, recurse=True, qmake="qmake", make="make", cache_dir=None):
	if cache_dir is not None:
		os.environ["QDEP_CACHE_DIR"] = cache_dir

	# eval the pro files if needed
	if evaluate:
		# special case: since we only download sources, a simple qmake run is enough, no actual "evaluation" needed
		for pro_file in targets:
			eval_pro_depends(pro_file, qmake, make)
			print("Done! qmake finished successfully, all qdep sources for the project tree have been downloaded.")
		return
	elif extract:
		packages = []
		for pro_file in targets:
			print("Extracting dependencies from {}...".format(pro_file))
			packages = packages + extract_pro_depends(pro_file, qmake)
	else:
		packages = list(targets)

	# download the actual sources
	print("Found {} initial packages".format(len(packages)))
	pkg_hashes = set()
	for package in packages:
		pkg_url, pkg_version, pkg_path = package_resolve(package)
		p_hash = pkg_hash(pkg_url, pkg_path)
		if p_hash in pkg_hashes:
			continue

		print("Downloading sources for {}...".format(package))
		cache_dir, _b = get_sources(pkg_url, pkg_version)

		if recurse:
			print("Extracting dependencies from {}...".format(package))
			new_packages = extract_pro_depends(path.join(cache_dir, pkg_path[1:]), qmake)
			print("Found {} dependent packages".format(len(new_packages)))
			for pkg in new_packages:
				packages.append(pkg)

		pkg_hashes.add(p_hash)
	print("Done!")


def update(pro_file, evaluate=False, replace=False, qmake="qmake", make="make"):
	if evaluate:
		all_deps = invert_map(eval_pro_depends(pro_file, qmake, make, dump_depends=True))
		pkg_all, pkg_new = check_for_updates(all_deps.keys())
		update_files = set()
		for old_pkg, new_pkg in pkg_new.items():
			for dep in all_deps[old_pkg]:
				update_files.add(dep)
			all_deps[new_pkg] = (all_deps[new_pkg] if new_pkg in all_deps else []) + all_deps[old_pkg]
			all_deps.pop(old_pkg)
		all_deps = invert_map(all_deps)
		for upd_file in update_files:
			replace_or_print_update(upd_file, all_deps[upd_file], pkg_new, replace)
	else:
		print("Extracting dependencies from {}...".format(pro_file))
		packages = extract_pro_depends(pro_file, qmake)
		pkg_all, pkg_new = check_for_updates(packages)
		replace_or_print_update(pro_file, pkg_all, pkg_new, replace)



from qdep.internal.common import *


def dephash(*packages, project=False, pkgpath=False):
	for package in packages:
		pkg_url, pkg_branch, pkg_path = package_resolve(package, project=project)
		if pkgpath:
			print(pkg_hash(pkg_url, pkg_path) + ";" + pkg_path)
		else:
			print(pkg_hash(pkg_url, pkg_path))


def pkgresolve(package, latest_version=None, project=False, pull=True, clone=True):
	ov_map = get_override_map()

	pkg_url, pkg_branch, pkg_path = package_resolve(package, pkg_version=latest_version, project=project)
	needs_cache = pkg_branch is None
	if pkg_url in ov_map:
		pkg_base = ov_map[pkg_url]
	else:
		pkg_base, pkg_branch = get_sources(pkg_url, pkg_branch, pull=pull, clone=clone)
	print(pkg_hash(pkg_url, pkg_path))
	print(pkg_branch)
	print(pkg_base)
	print(pkg_path)
	print(needs_cache)


def hookgen(prefix, header, resources=None, hooks=None):
	if resources is None:
		resources = []
	if hooks is None:
		hooks = []

	inc_guard = cpp_escape(path.basename(header)).upper()
	with open(header, "w") as out_file:
		out_file.write("#ifndef {}\n".format(inc_guard))
		out_file.write("#define {}\n\n".format(inc_guard))

		out_file.write("#include <QtCore/qglobal.h>\n\n")

		for hook in hooks:
			out_file.write(declare_hook(hook))

		out_file.write("\ninline void qdep_{}_init() {{\n".format(cpp_escape(prefix)))
		out_file.write("\t// resources\n")
		for resource in resources:  # ignore the first argument, as it is always the current pro file (to ensure this rule is always run
			out_file.write("\tQ_INIT_RESOURCE({});\n".format(cpp_escape(path.splitext(path.basename(resource))[0])))
		out_file.write("\t// hooks\n")
		for hook in hooks:
			out_file.write("\t::{}();\n".format(hook))
		out_file.write("}\n\n")

		out_file.write("#endif //{}\n".format(inc_guard))


def hookimp(outfile, headers=None, hooks=None):
	if headers is None:
		headers = []
	if hooks is None:
		hooks = []

	with open(outfile, "w") as out_file:
		targets = []
		target_regex = re.compile(r".*qdep_(.*)_hooks\.h$")

		out_file.write("#include <QtCore/QCoreApplication>\n")
		for header in headers:
			abs_path = path.abspath(header)
			out_file.write("#include \"{}\"\n".format(abs_path))
			targets.append(re.match(target_regex, header).group(1))

		out_file.write("\n")
		for hook in hooks:
			out_file.write(declare_hook(hook))

		out_file.write("\nnamespace {\n\n")
		out_file.write("void __qdep_startup_hooks() {\n")
		for target in targets:
			out_file.write("\tqdep_{}_init();\n".format(cpp_escape(target)))
		for hook in hooks:
			out_file.write("\t::{}();\n".format(hook))
		out_file.write("}\n\n")
		out_file.write("}\n")
		out_file.write("Q_COREAPP_STARTUP_FUNCTION(__qdep_startup_hooks)\n")


def lconvert(tsfile, outfile, *combine_files, lconvert_args=None):
	if lconvert_args is None:
		lconvert_args = ["lconvert"]

	# sort combine args into a map of the languages
	sub_map = dict()
	for sub_arg in combine_files:
		sub_base = path.splitext(path.basename(sub_arg))[0]
		sub_args = sub_base.split("_")[1:]
		for arg_cnt in range(len(sub_args)):
			suffix = "_".join(sub_args[arg_cnt:])
			if suffix not in sub_map:
				sub_map[suffix] = []
			sub_map[suffix].append(sub_arg)

	# find the qm files to combine with the input
	target_base = path.splitext(path.basename(tsfile))[0]
	ts_args = target_base.split("_")[1:]
	combine_list = []
	for arg_cnt in range(len(ts_args) - 1, -1, -1):
		suffix = "_".join(ts_args[arg_cnt:])
		if suffix in sub_map:
			combine_list = combine_list + sub_map[suffix]
	combine_list.append(tsfile)

	# run lconvert
	sub_run(lconvert_args + ["-if", "ts", "-i"] + combine_list + ["-of", "ts", "-o", outfile], check=True)


def prolink(prodir, pkghash, pkgpath, link=None):
	link_target = path.join(prodir, ".qdep", pkghash, "src")
	pro_target = path.join(link_target, pkgpath[1:])

	# skip mode -> only print the path
	if link is None:
		print(pro_target)
		return

	# normal mode -> first unlink any existing stuff
	if path.islink(link_target):
		if os.name == 'nt' and path.isdir(link_target):
			os.rmdir(link_target)
		else:
			os.unlink(link_target)
	elif path.isdir(link_target):
		shutil.rmtree(link_target)
	elif path.exists(link_target):
		os.remove(link_target)

	# then link the correct project again
	try:
		os.makedirs(path.dirname(link_target), exist_ok=True)
		os.symlink(link, link_target, target_is_directory=True)
	except OSError:
		# symlink is not possible, copy instead
		print("Project WARNING: Failed to symlink project dependecy. Performing deep copy instead", file=sys.stderr)
		shutil.copytree(link, link_target, symlinks=True, ignore_dangling_symlinks=True)

	print(pro_target)



import argparse

import argcomplete

from qdep.qdep import *
from qdep.internal.private import *


def complete_path(prefix, filter_fn):
	def map_dir_contents(base_dir, file):
		full_path = path.join(base_dir, file)
		if path.isdir(full_path):
			return full_path + "/"
		else:
			return full_path

	path_base, path_name = path.split(prefix)
	if len(path_base) == 0:
		path_base = "."
	return map(lambda f: map_dir_contents(path_base, f), filter(lambda f: filter_fn(path_base, path_name, f), os.listdir(path_base)))


def complete_executable(prefix, limiter=None):
	def filter_dir_contents(base_dir, base_name, file):
		if path.isdir(path.join(base_dir, file)):
			return True
		elif file.startswith(base_name) and (limiter is None or limiter in file):
			return True
		else:
			return False

	return complete_path(prefix, filter_dir_contents)


def complete_suffix(prefix, suffix):
	def filter_dir_contents(base_dir, base_name, file):
		if path.isdir(path.join(base_dir, file)):
			return True
		elif file.startswith(base_name) and file.endswith(suffix):
			return True
		else:
			return False

	return complete_path(prefix, filter_dir_contents)


def qmake_completer(prefix, **kwargs):
	return complete_executable(prefix, "qmake")


def make_completer(prefix, **kwargs):
	return complete_executable(prefix, "make")


def dir_completer(prefix, **kwargs):
	def filter_dir_contents(base_dir, base_name, file):
		return path.isdir(path.join(base_dir, file)) and file.startswith(base_name)

	return complete_path(prefix, filter_dir_contents)


def pro_completer(prefix, **kwargs):
	return complete_suffix(prefix, ".pro")


def pri_completer(prefix, **kwargs):
	return complete_suffix(prefix, ".pri")


def main():
	parser = argparse.ArgumentParser(description="A very basic yet simple to use dependency management tool for qmake based projects.")
	parser.add_argument("--version", action="version", version=version)
	parser.add_argument("--trace", action="store_true", help="In case of an exception, print the whole stack trace")

	sub_args = parser.add_subparsers(dest="operation", title="Operations", metavar="{operation}")

	prfgen_parser = sub_args.add_parser("prfgen", help="Generate a qmake project feature (prf) for the given qmake.")
	prfgen_parser.add_argument("--qmake", action="store", default="qmake", help="The path to a qmake executable to place the prf file for.").completer = qmake_completer
	prfgen_parser.add_argument("-d", "--dir", dest="dir", action="store", help="The directory containing the mkspec folder where to place the prf file. If not specified, qmake is queried form the location.").completer = dir_completer

	init_parser = sub_args.add_parser("init", help="Initialize a pro file to use qdep by adding the required lines.")
	init_parser.add_argument("profile", help="The path to the pro file to add the qdep code to.").completer = pro_completer

	lupdate_parser = sub_args.add_parser("lupdate", help="Run lupdate for the QDEP_TRANSLATION variable in a given pri file.")
	lupdate_parser.add_argument("--qmake", action="store", default="qmake", help="The path to a qmake executable to find the corresponding lupdate for.").completer = qmake_completer
	lupdate_parser.add_argument("--pri-file", dest="pri_path", action="store", required=True, help="The path to the pri-file that contains a QDEP_TRANSLATIONS variable, to generate translations for.").completer = pri_completer
	lupdate_parser.add_argument("largs", action="store", nargs="*", metavar="lupdate-argument", help="Additionals arguments to be passed to lupdate. MUST be proceeded by '--'!")

	clear_parser = sub_args.add_parser("clear", help="Remove all sources from the users global cache.")
	clear_parser.add_argument("-y", "--yes", dest="yes", action="store_true", help="Immediatly remove the caches, without asking for confirmation first.")

	versions_parser = sub_args.add_parser("versions", help="List all known versions/tags of the given package")
	versions_parser.add_argument("-b", "--branches", dest="branches", action="store_true", help="Include branches into the output.")
	versions_parser.add_argument("--no-tags", dest="tags", action="store_false", help="Exclude tags from the output.")
	versions_parser.add_argument("-s", "--short", dest="short", action="store_true", help="Print output as a single, uncommented line")
	versions_parser.add_argument("--limit", action="store", type=int, help="Limit the returned lists to the LIMIT newest entries per type.")
	versions_parser.add_argument("package", help="The package to list the versions for. Specify without a version of pro/pri file path!")

	query_parser = sub_args.add_parser("query", help="Query details about a given package identifier")
	query_parser.add_argument("--expand", action="store_true", help="Only expand the package name, don't output anything else.")
	query_parser.add_argument("--no-check", dest="check", action="store_false", help="Do not check if the package actually exists.")
	query_parser.add_argument("--versions", action="store_true", help="Also query and display all available tags and branches. See 'qdep.py versions' for alternative formats.")
	query_parser.add_argument("package", help="The package to query information for.")

	get_parser = sub_args.add_parser("get", help="Download the sources of one ore more packages into the source cache.")
	get_parser.add_argument("--extract", action="store_true", help="Run in pro-file mode. Arguments are interpreted as pro files and are scanned for dependencies")
	get_parser.add_argument("--eval", action="store_true", help="Fully evaluate all pro files by running qmake on them. Implies '--extract'.")
	get_parser.add_argument("--no-recurse", dest="recurse", action="store_false", help="Do not scan downloaded packages for further dependencies.")
	get_parser.add_argument("--qmake", action="store", default="qmake", help="The path to a qmake executable to use for evaluation if '--eval' was specified.").completer = qmake_completer
	get_parser.add_argument("--make", action="store", default="make", help="The path to a make executable to use for evaluation if '--eval' was specified.").completer = make_completer
	get_parser.add_argument("-d", "--dir", "--cache-dir", dest="dir", action="store", help="Specify the directory where to download the sources to. Shorthand for using the QDEP_CACHE_DIR environment variable.").completer = dir_completer
	get_parser.add_argument("args", nargs="+", help="The packages (or pro files if using '--extract') to download the sources for.").completer = pro_completer

	update_parser = sub_args.add_parser("update", help="Check for newer versions of used packages and optionally update them.")
	update_parser.add_argument("--eval", action="store_true", help="Fully evaluate all pro files by running qmake on them.")
	update_parser.add_argument("--qmake", action="store", default="qmake", help="The path to a qmake executable to use for evaluation if '--eval' was specified.").completer = qmake_completer
	update_parser.add_argument("--make", action="store", default="make", help="The path to a make executable to use for evaluation if '--eval' was specified.").completer = make_completer
	update_parser.add_argument("--replace", action="store_true", help="Automatically replace newer packages in the evaluated project files instead of printing to the console.")
	update_parser.add_argument("profile", metavar="pro-file", help="The qmake pro-file to update dependencies for.").completer = pro_completer

	dephash_parser = sub_args.add_parser("dephash", help="[INTERNAL] Generated unique identifying hashes for qdep packages.")
	dephash_parser.add_argument("--project", action="store_true", help="Interpret input as a project dependency, not a normal pri dependency.")
	dephash_parser.add_argument("--pkgpath", action="store_true", help="Return the hash and the pro/pri subpath as tuple, seperated by a ';'.")
	dephash_parser.add_argument("input", action="store", nargs="*", metavar="package", help="The packages to generate hashes for.")

	pkgresolve_parser = sub_args.add_parser("pkgresolve", help="[INTERNAL] Download the given qdep package and extract relevant information from it.")
	pkgresolve_parser.add_argument("--no-pull", dest="pull", action="store_false", help="Do not update existing packages that are based on branches instead of tags.")
	pkgresolve_parser.add_argument("--no-clone", dest="clone", action="store_false", help="Do not allow installation of new packages. Trying so will lead to an error. Updating existing packages is still possible.")
	pkgresolve_parser.add_argument("--project", dest="project", action="store_true", help="Interpret input as a project dependency, not a normal pri dependency")
	pkgresolve_parser.add_argument("package", action="store", metavar="package", help="The package identifier of the package to be downloaded and resolved.")
	pkgresolve_parser.add_argument("version", action="store", nargs="?", metavar="latest-version", help="The previously cached version for packages with no version identifier.")

	hookgen_parser = sub_args.add_parser("hookgen", help="[INTERNAL] Generate a header file with a method to load all resource hooks.")
	hookgen_parser.add_argument("--hooks", action="store", nargs="*", help="The names of additional hook functions to be referenced.")
	hookgen_parser.add_argument("prefix", action="store", help="The target name to use as part of the generated hook method.")
	hookgen_parser.add_argument("header", action="store", help="The path to the header-file to be generated.")
	hookgen_parser.add_argument("dummy", action="store", metavar="pro-file", help="The path to the current pro file - needed for Makefile rules.")
	hookgen_parser.add_argument("resources", action="store", nargs="*", metavar="resource", help="Paths to the resource-files to generate the hooks for.")

	hookimp_parser = sub_args.add_parser("hookimp", help="[INTERNAL] Generate a source file that includes and runs all qdep hooks as normal startup hook.")
	hookimp_parser.add_argument("--hooks", action="store", nargs="*", help="The names of additional hook functions to be referenced.")
	hookimp_parser.add_argument("outfile", action="store", help="The path to the cpp-file to be generated.")
	hookimp_parser.add_argument("dummy", action="store", metavar="pro-file", help="The path to the current pro file - needed for Makefile rules.")
	hookimp_parser.add_argument("headers", action="store", nargs="*", metavar="header", help="Paths to the header-files that contain hooks to be run.")

	lconvert_parser = sub_args.add_parser("lconvert", help="[INTERNAL] Combine ts files with translations from qdep packages.")
	lconvert_parser.add_argument("--combine", action="store", nargs="*", help="The qdep ts files that should be combined into the real ones.")
	lconvert_parser.add_argument("tsfile", action="store", help="The path to the ts file to combine with the qdep ts files.")
	lconvert_parser.add_argument("outfile", action="store", help="The path to the ts file to be generated.")
	lconvert_parser.add_argument("largs", action="store", nargs="+", metavar="lconvert-tool", help="Path to the lconvert tool as well as additional arguments to it.")

	prolink_parser = sub_args.add_parser("prolink", help="[INTERNAL] Resolve the path a linked project dependency would be at.")
	prolink_parser.add_argument("prodir", action="store", help="The directory of the pro file that includes the other one.")
	prolink_parser.add_argument("pkghash", action="store", help="The hash identifier of the project to link.")
	prolink_parser.add_argument("pkgpath", action="store", help="The path to the pro file within the dependency.")
	prolink_parser.add_argument("--link", action="store", help="Perform the link operation and create the symlink/dirtree, based on the given path to the dependency sources.")

	argcomplete.autocomplete(parser)
	res = parser.parse_args()

	try:
		if res.operation == "prfgen":
			prfgen(path.abspath(sys.argv[0]), res.qmake, res.dir)
		elif res.operation == "init":
			init(res.profile)
		elif res.operation == "lupdate":
			lupdate(res.pri_path, res.qmake, res.largs)
		elif res.operation == "clear":
			clear(res.yes)
		elif res.operation == "versions":
			versions(res.package, res.tags, res.branches, res.short, res.limit)
		elif res.operation == "query":
			query(res.package, res.check, res.versions, res.expand)
		elif res.operation == "get":
			get(*res.args, extract=res.extract, evaluate=res.eval, recurse=res.recurse, qmake=res.qmake, make=res.make, cache_dir=res.dir)
		elif res.operation == "update":
			update(res.profile, res.eval, res.replace, res.qmake, res.make)
		elif res.operation == "dephash":
			dephash(*res.input, project=res.project, pkgpath=res.pkgpath)
		elif res.operation == "pkgresolve":
			pkgresolve(res.package, res.version, res.project, res.pull, res.clone)
		elif res.operation == "hookgen":
			hookgen(res.prefix, res.header, res.resources, res.hooks)
		elif res.operation == "hookimp":
			hookimp(res.outfile, res.headers, res.hooks)
		elif res.operation == "lconvert":
			lconvert(res.tsfile, res.outfile, *res.combine, lconvert_args=res.largs)
		elif res.operation == "prolink":
			prolink(res.prodir, res.pkghash, res.pkgpath, res.link)
		else:
			parser.print_help()
			return 1

		return 0
	except Exception as exception:
		if res.trace or True:
			raise
		else:
			print(exception, file=sys.stderr)
		return 1


if __name__ == '__main__':
	sys.exit(main())



qdep_prf = """isEmpty(QDEP_TOOL) QDEP_TOOL = $$system_path($$QDEP_PATH)

# verify versions are correct
__qdep_script_version = $$system($$QDEP_TOOL --version)
!equals(QDEP_VERSION, $$__qdep_script_version):error("qdep.py script version ($$__qdep_script_version) is different to qdep.prf version ($$QDEP_VERSION)! Run '$$QDEP_TOOL prfgen --qmake $$QMAKE_QMAKE' to update the prf file!")

# set some variables
isEmpty(QDEP_CACHE_SCOPE): QDEP_CACHE_SCOPE = stash

isEmpty(QDEP_GENERATED_DIR): QDEP_GENERATED_DIR = $$OUT_PWD
debug_and_release:CONFIG(release, debug|release): QDEP_GENERATED_SOURCES_DIR = $${QDEP_GENERATED_DIR}/release
else:debug_and_release:CONFIG(debug, debug|release): QDEP_GENERATED_SOURCES_DIR = $${QDEP_GENERATED_DIR}/debug
else: QDEP_GENERATED_SOURCES_DIR = $$QDEP_GENERATED_DIR
isEmpty(QDEP_GENERATED_TS_DIR): QDEP_GENERATED_TS_DIR = $$QDEP_GENERATED_DIR/.qdepts
debug_and_release:CONFIG(release, debug|release): QDEP_GENERATED_TS_DIR = $${QDEP_GENERATED_TS_DIR}/release
else:debug_and_release:CONFIG(debug, debug|release): QDEP_GENERATED_TS_DIR = $${QDEP_GENERATED_TS_DIR}/debug
isEmpty(QDEP_LUPDATE) {
	qtPrepareTool(QDEP_LUPDATE, lupdate)
	QDEP_LUPDATE += -recursive -locations relative
    qdep_lupdate_no_obsolete: QDEP_LUPDATE += -no-obsolete
}
isEmpty(QDEP_LCONVERT) {
	qtPrepareTool(QDEP_LCONVERT, lconvert)
	QDEP_LCONVERT += -sort-contexts 
}

isEmpty(QDEP_EXPORT_PATH): QDEP_EXPORT_PATH = $$QDEP_GENERATED_DIR
debug_and_release:CONFIG(release, debug|release): QDEP_EXPORT_PATH = $$QDEP_EXPORT_PATH/release
debug_and_release:CONFIG(debug, debug|release): QDEP_EXPORT_PATH = $$QDEP_EXPORT_PATH/debug
isEmpty(QDEP_EXPORT_NAME) {
	__qdep_base_name = $$basename(_PRO_FILE_)
	QDEP_EXPORT_NAME = $$replace(__qdep_base_name, "\\\\.[^\\\\.]*$", "_export.pri")
}

isEmpty(__QDEP_PRIVATE_SEPERATOR): __QDEP_PRIVATE_SEPERATOR = "==="
isEmpty(__QDEP_TUPLE_SEPERATOR): __QDEP_TUPLE_SEPERATOR = "---"

CONFIG += qdep_build
DEFINES += QDEP_BUILD

# The primary dependecy collector function
defineTest(qdepCollectDependencies) {
	# transform all dependencies into unique hashes
	qdep_dependencies = 
	for(arg, ARGS): qdep_dependencies += $$system_quote($$arg)
	qdep_ok = 
	qdep_hashes = $$system($$QDEP_TOOL dephash $$qdep_dependencies, lines, qdep_ok)
	!equals(qdep_ok, 0):return(false)

	for(dep_hash, qdep_hashes) {
		# handle each dependency, but each package only once
		dep_pkg = $$take_first(ARGS)
		!contains(__QDEP_INCLUDE_CACHE, $$dep_hash) {
			# Install the sources and extract some parameters
			dep_extra_args = 
			qdep_no_pull: dep_extra_args += --no-pull
			qdep_no_clone: dep_extra_args += --no-clone
			!isEmpty($${dep_hash}.version): dep_vers_arg = $$system_quote($$first($${dep_hash}.version))
			else: dep_vers_arg =
			qdep_ok = 
			dep_data = $$system($$QDEP_TOOL pkgresolve $$dep_extra_args $$system_quote($$dep_pkg) $$dep_vers_arg, lines, qdep_ok)
			!equals(qdep_ok, 0):return(false)
			!equals(dep_hash, $$take_first(dep_data)):error("Critical internal error: dependencies out of sync"):return(false)

			dep_version = $$take_first(dep_data)
			dep_base = $$take_first(dep_data)
			dep_path = $$take_first(dep_data)
			dep_needs_cache = $$take_first(dep_data)

			$${dep_hash}.package = $$dep_pkg
			$${dep_hash}.version = $$dep_version
			$${dep_hash}.path = $${dep_base}$${dep_path}
			$${dep_hash}.exports = 
			$${dep_hash}.local = 1
			export($${dep_hash}.package)
			export($${dep_hash}.version)
			export($${dep_hash}.path)
			export($${dep_hash}.local)

			# Cache the package version for dependencies with undetermined versions
			!qdep_no_cache:equals(dep_needs_cache, True):!cache($${dep_hash}.version, set $$QDEP_CACHE_SCOPE):warning("Failed to cache package version for $$dep_pkg")

			# Find all dependencies that package depends on and call this method recursively for those
			sub_deps = $$fromfile($$eval($${dep_hash}.path), QDEP_DEPENDS)
			__QDEP_REAL_DEPS_STACK += $$dep_hash			
			!isEmpty(sub_deps):!qdepCollectDependencies($$sub_deps):return(false)			
			__QDEP_INCLUDE_CACHE *= $$take_last(__QDEP_REAL_DEPS_STACK)
			export(__QDEP_INCLUDE_CACHE)

			# Handle all defines for symbol exports, if specified
			sub_exports = $$fromfile($$eval($${dep_hash}.path), QDEP_PACKAGE_EXPORTS)

			qdep_export_all|contains(QDEP_EXPORTS, $$dep_pkg) {
				!static:!staticlib:for(sub_export, sub_exports) {
					DEFINES += "$${sub_export}=Q_DECL_EXPORT"
					QDEP_EXPORTED_DEFINES += "$${sub_export}=Q_DECL_IMPORT"
					$${dep_hash}.exports += $$sub_export
				} else:for(sub_export, sub_exports) {
					DEFINES += "$${sub_export}="
					QDEP_EXPORTED_DEFINES += "$${sub_export}="
				}
			} else:for(sub_export, sub_exports): \\
				DEFINES += "$${sub_export}="
			export(DEFINES)
			export(QDEP_EXPORTED_DEFINES)
			export($${dep_hash}.exports)
		} else: \\
			!equals($${dep_hash}.package, $$dep_pkg): \\
			warning("Detected includes of multiple different versions of the same dependency. Package \\"$$first($${dep_hash}.package)\\" is used, and package \\"$$dep_pkg\\" was detected.")
	}

	return(true)
}

# The dependecy collector function for project dependencies in subdirs projects
defineTest(qdepCollectProjectDependencies) {
	# transform all dependencies into unique hashes
	qdep_dependencies = 
	for(arg, ARGS): qdep_dependencies += $$system_quote($$arg)
	qdep_ok = 
	qdep_hashes = $$system($$QDEP_TOOL dephash --project $$qdep_dependencies, lines, qdep_ok)
	!equals(qdep_ok, 0):return(false)

	for(dep_hash, qdep_hashes) {
		# handle each dependency, but each package only once
		dep_pkg = $$take_first(ARGS)
		!contains(__QDEP_INCLUDE_CACHE, $$dep_hash) {
			# Install the sources and extract some parameters
			dep_extra_args = 
			qdep_no_pull: dep_extra_args += --no-pull
			qdep_no_clone: dep_extra_args += --no-clone
			!isEmpty($${dep_hash}.version): dep_vers_arg = $$system_quote($$first($${dep_hash}.version))
			else: dep_vers_arg =
			qdep_ok = 
			dep_data = $$system($$QDEP_TOOL pkgresolve --project $$dep_extra_args $$system_quote($$dep_pkg) $$dep_vers_arg, lines, qdep_ok)
			!equals(qdep_ok, 0):return(false)
			!equals(dep_hash, $$take_first(dep_data)):error("Critical internal error: project dependencies out of sync"):return(false)

			dep_version = $$take_first(dep_data)
			dep_base = $$take_first(dep_data)
			dep_path = $$take_first(dep_data)
			dep_needs_cache = $$take_first(dep_data)

			$${dep_hash}.package = $$dep_pkg
			$${dep_hash}.version = $$dep_version
			$${dep_hash}.path = $${dep_base}$${dep_path}
			$${dep_hash}.exports = 
			$${dep_hash}.local = 0
			$${dep_hash}.target = sub$${dep_hash}
			export($${dep_hash}.package)
			export($${dep_hash}.version)
			export($${dep_hash}.path)
			export($${dep_hash}.local)
			export($${dep_hash}.target)

			# Cache the package version for dependencies with undetermined versions
			!qdep_no_cache:equals(dep_needs_cache, True):!cache($${dep_hash}.version, set $$QDEP_CACHE_SCOPE):warning("Failed to cache package version for $$dep_pkg")

			# link the project and write the version if not already there
			exists("$$_PRO_FILE_PWD_/.qdep/$${dep_hash}/.version"): linked_version = $$cat("$$_PRO_FILE_PWD_/.qdep/$${dep_hash}/.version", lines)
			else: linked_version =			
			qdep_ok = 
			!equals($${dep_hash}.version, $$linked_version) {
				$${dep_hash}.file = $$system($$QDEP_TOOL prolink --link $$system_quote($${dep_base}) $$system_quote($$_PRO_FILE_PWD_) $${dep_hash} $$system_quote($${dep_path}), lines, qdep_ok)
				!equals(qdep_ok, 0):return(false)
				!write_file("$$_PRO_FILE_PWD_/.qdep/$${dep_hash}/.version", $${dep_hash}.version): \\
					error("Failed to cache project dependency version for $$first($${dep_hash}.package)")
			} else {
				$${dep_hash}.file = $$system($$QDEP_TOOL prolink $$system_quote($$_PRO_FILE_PWD_) $${dep_hash} $$system_quote($${dep_path}), lines, qdep_ok)
				!equals(qdep_ok, 0):return(false)
			}
			export($${dep_hash}.file)

			# Find all further project dependencies that package depends on
			sub_deps = $$fromfile($$eval($${dep_hash}.path), QDEP_PROJECT_DEPENDS)
			!isEmpty(sub_deps) {
				qdep_ok = 
				$${dep_hash}.depends = $$system($$QDEP_TOOL dephash --project $$sub_deps, lines, qdep_ok)
				!equals(qdep_ok, 0):return(false)
				export($${dep_hash}.depends)
			}
			SUBDIRS += $${dep_hash}
			export(SUBDIRS)
			__QDEP_INCLUDE_CACHE += $${dep_hash}
			export(__QDEP_INCLUDE_CACHE)

			# calls this method recursively for all dependencies
			!qdepCollectProjectDependencies($$sub_deps):return(false)	
		} else: \\
			!equals($${dep_hash}.package, $$dep_pkg): \\
			warning("Detected includes of multiple different versions of the same dependency. Package \\"$$first($${dep_hash}.package)\\" is used, and package \\"$$dep_pkg\\" was detected.")
	}

	return(true)
}

# resolve qdep_depends of subdir dependencies to the actual hashes
defineTest(qdepResolveSubdirDepends) {
	for(subproj, 1) {
		!isEmpty($${subproj}.qdep_depends) {
			qdep_dependencies = 
			for(arg, $${subproj}.qdep_depends): qdep_dependencies += $$system_quote($$arg)
			qdep_ok = 
			$${subproj}.depends += $$system($$QDEP_TOOL dephash --project $$qdep_dependencies, lines, qdep_ok)
			!equals(qdep_ok, 0):return(false)
			export($${subproj}.depends)
		}
	}
	return(true)
}

# pass the root dir and dependencies and create QDEP_LINK_DEPENDS values from it
defineReplace(qdepResolveProjectLinkDeps) {
	qdep_dependencies = 
	for(arg, 2): qdep_dependencies += $$system_quote($$arg)
	qdep_ok = 
	dep_tuples = $$system($$QDEP_TOOL dephash --project --pkgpath $$qdep_dependencies, lines, qdep_ok)
	!equals(qdep_ok, 0):return()

	link_paths = 
	for(qdep_tuple, dep_tuples) {
		tpl_args = $$split(qdep_tuple, ";")
		dep_hash = $$take_first(tpl_args)
		dep_path = $$join(tpl_args, ";")
		qdep_ok = 
		link_paths += $$system($$QDEP_TOOL prolink $$system_quote($$absolute_path($$1, $$_PRO_FILE_PWD_)) $${dep_hash} $$system_quote($${dep_path}), lines, qdep_ok)
		!equals(qdep_ok, 0):return()
	}

	return($$link_paths)
}

# Write a quoted value for the given variable name as a single value
defineReplace(qdepOutQuote) {
	result = 
	var_name = $$1
	isEmpty(4): \\
		intendent = $$escape_expand(\\t)
	else {
		intendent = 
		for(x, "1..$$4"): \\
			intendent = "$$intendent$$escape_expand(\\t)"
	}
	equals(3, prepend): \\
		for(value, 2): result += "$$intendent$$var_name = $${LITERAL_DOLLAR}$${LITERAL_DOLLAR}quote($$value) $${LITERAL_DOLLAR}$${LITERAL_DOLLAR}$$var_name"
	else:equals(3, star): \\
		for(value, 2): result += "$$intendent$$var_name *= $${LITERAL_DOLLAR}$${LITERAL_DOLLAR}quote($$value)"
	else:equals(3, pop): \\
		for(value, 2): result += "$$intendent$${LITERAL_DOLLAR}$${LITERAL_DOLLAR}take_last($$var_name)"
	else: \\
		for(value, 2): result += "$$intendent$$var_name += $${LITERAL_DOLLAR}$${LITERAL_DOLLAR}quote($$value)"
	return($$result)
}

# A function to create a pri file to include the library and all exported
defineTest(qdepCreateExportPri) {
	# write include guard
	out_file_data = "!contains(__QDEP_EXPORT_QMAKE_INCLUDE_GUARD, $${LITERAL_DOLLAR}$${LITERAL_DOLLAR}PWD) {"
	out_file_data += $$qdepOutQuote(__QDEP_EXPORT_QMAKE_INCLUDE_GUARD, $${LITERAL_DOLLAR}$${LITERAL_DOLLAR}PWD)
	
	# write dependencies
	!static:!staticlib: \\
		out_file_data += $$qdepOutQuote(__QDEP_IS_DYNAMIC_EXPORT_INCLUDE_STACK, true)
	for(link_dep, QDEP_LINK_DEPENDS): \\
		out_file_data += "$$escape_expand(\\t)include($$qdepLinkExpand($$link_dep))"
	!static:!staticlib: \\
		out_file_data += $$qdepOutQuote(__QDEP_IS_DYNAMIC_EXPORT_INCLUDE_STACK, true, pop)
	
	# write basic variables
	out_file_data += $$qdepOutQuote(DEFINES, $$QDEP_EXPORTED_DEFINES $$QDEP_DEFINES)
	out_file_data += $$qdepOutQuote(INCLUDEPATH, $$QDEP_EXPORTED_INCLUDEPATH $$QDEP_INCLUDEPATH)
	out_file_data += $$qdepOutQuote(LIBS, $$QDEP_EXPORTED_LIBS $$QDEP_LIBS)
	for(exp_var_key, QDEP_VAR_EXPORTS): out_file_data += $$qdepOutQuote($$exp_var_key, $$eval($$exp_var_key))

	# write package cache
	for(dep_hash, __QDEP_INCLUDE_CACHE):equals($${dep_hash}.local, 1) {
		out_file_data += $$qdepOutQuote($${dep_hash}.package, $$eval($${dep_hash}.package))
		out_file_data += $$qdepOutQuote($${dep_hash}.version, $$eval($${dep_hash}.version))
		out_file_data += $$qdepOutQuote($${dep_hash}.path, $$eval($${dep_hash}.path))
		out_file_data += $$qdepOutQuote($${dep_hash}.exports, $$eval($${dep_hash}.exports))
		out_file_data += $$qdepOutQuote($${dep_hash}.local, 0)
		out_file_data += $$qdepOutQuote(__QDEP_INCLUDE_CACHE, $$dep_hash)
	}

	# write library linkage common parts
	isEmpty(DESTDIR) {
		out_libdir = $$OUT_PWD
		debug_and_release:CONFIG(release, debug|release): out_libdir = $${out_libdir}/release
		else:debug_and_release:CONFIG(debug, debug|release): out_libdir = $${out_libdir}/debug
	} else: out_libdir = $$absolute_path($$DESTDIR, $$OUT_PWD)
		
	!qdep_no_link {
		# write includepath
		out_file_data += $$qdepOutQuote(INCLUDEPATH, $$_PRO_FILE_PWD_)
	
		# write QT, PKGCONFIG and QDEP_LIBS (unless disabled)
		!qdep_no_export_link {
			qt: out_file_data += $$qdepOutQuote(QT, $$QT, star)
			link_pkgconfig: out_file_data += $$qdepOutQuote(PKGCONFIG, $$PKGCONFIG, star)
		}
	}

	# write static/dynamic specific library extra stuff
	static|staticlib:equals(TEMPLATE, lib) {
		out_file_data += "$$escape_expand(\\t)isEmpty(__QDEP_IS_DYNAMIC_EXPORT_INCLUDE_STACK) {"
		
		# write startup hooks
		debug_and_release:CONFIG(release, debug|release): out_file_data += "$$qdepOutQuote(__QDEP_HOOK_FILES, $$QDEP_GENERATED_DIR/release/qdep_$${TARGET}_hooks.h, "", 2)"
		else:debug_and_release:CONFIG(debug, debug|release): out_file_data += "$$qdepOutQuote(__QDEP_HOOK_FILES, $$QDEP_GENERATED_DIR/debug/qdep_$${TARGET}_hooks.h, "", 2)"
		else: out_file_data += "$$qdepOutQuote(__QDEP_HOOK_FILES, $$QDEP_GENERATED_DIR/qdep_$${TARGET}_hooks.h, "", 2)"
		
		# linkage
		!qdep_no_link {
			out_file_data += $$qdepOutQuote(DEPENDPATH, $$_PRO_FILE_PWD_, "", 2)

			win32-g++: out_file_data += $$qdepOutQuote(PRE_TARGETDEPS, "$${out_libdir}/lib$${TARGET}.a", "", 2)
			else:win32: out_file_data += $$qdepOutQuote(PRE_TARGETDEPS, "$${out_libdir}/$${TARGET}.lib", "", 2)
			else:unix: out_file_data += $$qdepOutQuote(PRE_TARGETDEPS, "$${out_libdir}/lib$${TARGET}.a", "", 2)

			out_file_data += $$qdepOutQuote(LIBS, "-l$${TARGET}", prepend, 2)
			out_file_data += $$qdepOutQuote(LIBS, "-L$${out_libdir}/", prepend, 2)
		}
		
		out_file_data += "$$escape_expand(\\t)}"
	} else {
		# linkage
		!qdep_no_link {
			equals(TEMPLATE, lib):out_file_data += $$qdepOutQuote(LIBS, "-l$${TARGET}", prepend)
			else {
				win32: bin_suffix = .exe
				out_file_data += $$qdepOutQuote(LIBS, "-l:$${TARGET}$${bin_suffix}", prepend)
			}
			out_file_data += $$qdepOutQuote(LIBS, "-L$${out_libdir}/", prepend)
		}
	}

	out_file_data += "}"
	write_file($$1, out_file_data):return(true)
	else:return(false)
}

# get the full pri path of a link dependency
defineReplace(qdepLinkExpand) {
	base_path = $$1
	suffix = $$str_member($$base_path, -4, -1)

	!equals(suffix, ".pri") {
		equals(suffix, ".pro") {
			file_name = $$basename(base_path)
			file_name = "$$str_member($$file_name, 0, -5)_export.pri"
			base_path = $$dirname(base_path)
		} else {
			file_name = "$$basename(base_path)_export.pri"
		}
		debug_and_release:CONFIG(release, debug|release): base_path = $$base_path/release/$$file_name
		else:debug_and_release:CONFIG(debug, debug|release):  base_path = $$base_path/debug/$$file_name
		else: base_path = $$base_path/$$file_name
	}

	base_path = $$absolute_path($$base_path, $$_PRO_FILE_PWD_)
	s_base_path = $$shadowed($$base_path)
	!isEmpty(s_base_path):exists($$s_base_path):return($$s_base_path)
	else:!isEmpty(base_path):exists($$base_path):return($$base_path)
	else:return()
}

# find the SUBDIRS project build directory that this qdep project was linked from
defineReplace(qdepResolveLinkRoot) {
	path_segments = $$split(1, "/")
	path_count = $$size(path_segments)
	c_path = $$1
	check_next = 0
	for(index, 0..$$path_count) {
		equals(check_next, 1): \\
			exists("$$c_path/Makefile"): \\
			return($$c_path)

		c_name = $$basename(c_path)
		equals(c_name, ".qdep"): check_next = 1
		else: check_next = 0
		c_path = $$dirname(c_path)
		!exists($$c_path):return()
	}
	return()
}

# shell_escape a list of files
defineReplace(qdepShellQuote) {
	out_files = 
	for(input, ARGS): out_files += $$shell_quote($$absolute_path($$input, $$_PRO_FILE_PWD_))
	return($$out_files)
}

# dump dependencies for update collection
defineTest(qdepDumpUpdateDeps) {
	dump_data = $$_PRO_FILE_
	dump_data += $$QDEP_DEPENDS
	dump_data += $$QDEP_PROJECT_SUBDIRS
	dump_data += $$QDEP_PROJECT_LINK_DEPENDS
	dump_data += $$QDEP_PROJECT_DEPENDS
	write_file($$OUT_PWD/qdep_depends.txt, dump_data):return(true)
	else:return(false)
}

# dump deps if in update mode
__qdep_dump_dependencies: \\
	!qdepDumpUpdateDeps(): \\
	error("Failed to dump dependencies")

# First transform project link depends to normal link depends
!isEmpty(QDEP_PROJECT_ROOT): \\
	!isEmpty(QDEP_PROJECT_LINK_DEPENDS): \\
	QDEP_LINK_DEPENDS = $$qdepResolveProjectLinkDeps($$QDEP_PROJECT_ROOT, $$QDEP_PROJECT_LINK_DEPENDS) $$QDEP_LINK_DEPENDS  # always prepend project link depends

# Second transform project depends to normal link depends
# This will only work if the project is imported as qdep project dependency
!isEmpty(QDEP_PROJECT_DEPENDS) {
	__qdep_project_link_root = $$qdepResolveLinkRoot($$OUT_PWD)
	isEmpty(__qdep_project_link_root): warning("Failed to find including subdirs project - only use QDEP_PROJECT_DEPENDS in qdep project depenencies")
	else: QDEP_LINK_DEPENDS = $$qdepResolveProjectLinkDeps($$__qdep_project_link_root, $$QDEP_PROJECT_DEPENDS) $$QDEP_LINK_DEPENDS  # always prepend project link depends
}

# Next collect all indirect dependencies
# for GCC: create a link group, as these all add libs in arbitrary order
!isEmpty(QDEP_LINK_DEPENDS): \\
	for(link_dep, QDEP_LINK_DEPENDS): \\
	!include($$qdepLinkExpand($$link_dep)): \\
	error("Failed to include linked library $$link_dep")

# Collect all dependencies and then include them
!isEmpty(QDEP_DEPENDS): {
	__qdep_cached_hooks = $$QDEP_HOOK_FNS
	QDEP_HOOK_FNS = 
	__qdep_cached_resources = $$RESOURCES
	RESOURCES = 

	!qdepCollectDependencies($$QDEP_DEPENDS):error("Failed to collect all dependencies")
	else:for(dep, __QDEP_INCLUDE_CACHE):equals($${dep}.local, 1) {
		__qdep_define_offset = $$size(DEFINES)
		__qdep_include_offset = $$size(INCLUDEPATH)
		__qdep_libs_offset = $$size(LIBS)
		include($$first($${dep}.path)) {
			qdep_export_all|contains(QDEP_EXPORTS, $$first($${dep}.package)) {
				QDEP_EXPORTED_DEFINES += $$member(DEFINES, $$__qdep_define_offset, -1)
				export(QDEP_EXPORTED_DEFINES)
				QDEP_EXPORTED_INCLUDEPATH += $$member(INCLUDEPATH, $$__qdep_include_offset, -1)
				export(QDEP_EXPORTED_INCLUDEPATH)
				QDEP_EXPORTED_LIBS += $$member(LIBS, $$__qdep_libs_offset, -1)
				export(QDEP_EXPORTED_LIBS)
			}
		} else:error("Failed to include pri file $$first($${dep}.package)")
	}

	QDEP_HOOK_FNS += $$__qdep_cached_hooks
	RESOURCES += $$__qdep_cached_resources
}

# Collect all project dependencies
equals(TEMPLATE, subdirs):!isEmpty(QDEP_PROJECT_SUBDIRS) {
	!qdepCollectProjectDependencies($$QDEP_PROJECT_SUBDIRS): \\
		error("Failed to collect all project dependencies")
	!qdepResolveSubdirDepends($$SUBDIRS): \\
		error("Failed to link all project dependencies")
}

# create special target for resource hooks in static libs
# or if not, reference their hooks of static libs (as special compiler, only for non-static apps/libs)
static|staticlib:equals(TEMPLATE, lib) {
	__qdep_hook_generator_c.name = qdep hookgen ${QMAKE_FILE_IN}
	__qdep_hook_generator_c.input = _PRO_FILE_ RESOURCES 
	__qdep_hook_generator_c.variable_out = HEADERS
	__qdep_hook_generator_c.commands = $$QDEP_PATH hookgen --hooks $$QDEP_HOOK_FNS -- $${TARGET} ${QMAKE_FILE_OUT} ${QMAKE_FILE_IN}
	__qdep_hook_generator_c.output = $$QDEP_GENERATED_SOURCES_DIR/qdep_$${TARGET}_hooks.h
	__qdep_hook_generator_c.CONFIG += target_predeps combine no_link
	__qdep_hook_generator_c.depends += $$QDEP_PATH
	QMAKE_EXTRA_COMPILERS += __qdep_hook_generator_c
} else:!equals(TEMPLATE, aux) {
	__qdep_hook_importer_c.name = qdep hookimp ${QMAKE_FILE_IN}
	__qdep_hook_importer_c.input = _PRO_FILE_ __QDEP_HOOK_FILES
	__qdep_hook_importer_c.variable_out = GENERATED_SOURCES
	__qdep_hook_importer_c.commands = $$QDEP_PATH hookimp --hooks $$QDEP_HOOK_FNS -- ${QMAKE_FILE_OUT} ${QMAKE_FILE_IN}
	__qdep_hook_importer_c.output = $$QDEP_GENERATED_SOURCES_DIR/qdep_imported_hooks.cpp
	__qdep_hook_importer_c.CONFIG += target_predeps combine
	__qdep_hook_importer_c.depends += $$QDEP_PATH
	QMAKE_EXTRA_COMPILERS += __qdep_hook_importer_c
}

# Create lupdate target
__qdep_lupdate_target.target = lupdate
__qdep_lupdate_target.commands = $$QDEP_LUPDATE $$qdepShellQuote($$_PRO_FILE_PWD_ $$QDEP_LUPDATE_INPUTS) -ts $$qdepShellQuote($$TRANSLATIONS)
__qdep_lupdate_target.depends += $$QDEP_LUPDATE_EXE $$_PRO_FILE_PWD_ $$QDEP_LUPDATE_INPUTS
QMAKE_EXTRA_TARGETS += __qdep_lupdate_target

# fix for broken lrelease make feature
{
	isEmpty(LRELEASE_DIR): LRELEASE_DIR = .qm
	debug_and_release:CONFIG(release, debug|release): __qdep_lrelease_real_dir = $${LRELEASE_DIR}/release
	else:debug_and_release:CONFIG(debug, debug|release): __qdep_lrelease_real_dir = $${LRELEASE_DIR}/debug
	else: __qdep_lrelease_real_dir = $$LRELEASE_DIR
	__qdep_lrelease_real_dir = $$absolute_path($$__qdep_lrelease_real_dir, $$OUT_PWD)
	!exists($$__qdep_lrelease_real_dir): \\
		!mkpath($$__qdep_lrelease_real_dir): \\
		warning("Failed to create lrelease directory: $$__qdep_lrelease_real_dir")
}
qm_files.CONFIG += no_check_exist

# Create special targets for translations
!qdep_no_qm_combine {
	# move translations into temporary var for the compiler to work
	__QDEP_ORIGINAL_TRANSLATIONS = $$TRANSLATIONS
	TRANSLATIONS = 

	# compiler for combined translations
	__qdep_qm_combine_c.name = qdep lrelease ${QMAKE_FILE_IN}
	__qdep_qm_combine_c.input = __QDEP_ORIGINAL_TRANSLATIONS 
	__qdep_qm_combine_c.variable_out = TRANSLATIONS
	__qdep_qm_combine_c.commands = $$QDEP_PATH lconvert --combine $$qdepShellQuote($$QDEP_TRANSLATIONS) -- ${QMAKE_FILE_IN} ${QMAKE_FILE_OUT} $$QDEP_LCONVERT
	__qdep_qm_combine_c.output = $$QDEP_GENERATED_TS_DIR/${QMAKE_FILE_BASE}.ts
	__qdep_qm_combine_c.CONFIG += no_link
	__qdep_qm_combine_c.depends += $$QDEP_PATH $$QDEP_LCONVERT_EXE $$QDEP_TRANSLATIONS
	QMAKE_EXTRA_COMPILERS += __qdep_qm_combine_c

	# copy from lrelease.prf - needed as TRANSLATIONS is now empty
	for(translation, $$list($$__QDEP_ORIGINAL_TRANSLATIONS $$EXTRA_TRANSLATIONS)) {
		translation = $$basename(translation)
		QM_FILES += $$__qdep_lrelease_real_dir/$$replace(translation, \\\\..*$, .qm)
	}
} else {
	# add qdep translations to extra to allow lrelease processing
	EXTRA_TRANSLATIONS += $$QDEP_TRANSLATIONS
}

# Create qdep pri export, if modules should be exported
equals(TEMPLATE, lib):!qdep_no_link|qdep_link_export|qdep_export_all|!isEmpty(QDEP_EXPORTS): \\
	!qdepCreateExportPri($$QDEP_EXPORT_PATH/$$QDEP_EXPORT_NAME): \\
	error("Failed to create export file $$QDEP_EXPORT_FILE")

DEFINES += $$QDEP_DEFINES
INCLUDEPATH += $$QDEP_INCLUDEPATH
LIBS += $$QDEP_LIBS
"""



import hashlib
import os
import re
import shutil
import stat
import subprocess
import sys
import tempfile
from os import path
from collections import OrderedDict

import appdirs
from lockfile import LockFile


def get_cache_dir_default():
	return appdirs.user_cache_dir("qdep")


def get_cache_dir(pkg_url, pkg_branch):
	cache_dir = os.getenv("QDEP_CACHE_DIR", get_cache_dir_default())
	cache_dir = path.join(cache_dir, hashlib.sha3_256(pkg_url.encode("UTF-8")).hexdigest(), pkg_branch)
	os.makedirs(cache_dir, exist_ok=True)
	return cache_dir


def get_override_map():
	or_env = os.getenv("QDEP_SOURCE_OVERRIDE")
	if or_env is None:
		return {}

	or_map = {}
	for env_info in or_env.split(";"):
		env_pair = env_info.split("^")
		if len(env_pair) == 2:
			or_map[env_pair[0]] = env_pair[1]
	return or_map


def pkg_hash(pkg_url, pkg_path):
	return "__QDEP_PKG_" + hashlib.sha3_256((pkg_url + pkg_path).lower().encode("UTF-8")).hexdigest()


def cpp_escape(name, ascii=True):
	return re.sub(r"[^\w]", "_", name, re.ASCII if ascii else re.UNICODE)


def declare_hook(hook):
	last_sep = hook.rfind("::")
	if last_sep != -1:
		return "namespace {} {{ void {}(); }}\n".format(hook[:last_sep], hook[last_sep + 2:])
	else:
		return "void {}();\n".format(hook)


def sub_run(*args, **kwargs):
	sys.stdout.flush()
	sys.stderr.flush()
	return subprocess.run(*args, **kwargs)


def package_resolve(package, pkg_version=None, project=False, expand=True):
	pattern = re.compile(r'^(?:([^@\/]+\/[^@\/]+)|(\w+:\/\/.*\.git|[^@:]*@[^@]*:[^@]+\.git))(?:@([^\/\s]+)(\/.*)?)?$')
	path_pattern = re.compile(r'^.*\/([^\/]+)\/?\.git$')

	match = re.match(pattern, package)
	if not match:
		raise Exception("Given package is not a valid package descriptor: " + package)

	# extract package url
	if match.group(1) is not None:
		if not expand:
			pkg_url = match.group(1)
		else:
			pkg_url = os.getenv("QDEP_DEFAULT_PKG_FN", "https://github.com/{}.git").format(match.group(1))
	elif match.group(2) is not None:
		pkg_url = match.group(2)
	else:
		raise Exception("Invalid package specified: " + package)

	# extract branch
	if match.group(3) is not None:
		pkg_branch = match.group(3)
	elif pkg_version is not None:
		pkg_branch = pkg_version
	else:
		pkg_branch = None

	# extract pri path
	if match.group(4) is not None:
		pkg_path = match.group(4)
	elif not expand:
		pkg_path = ""
	else:
		pkg_path = "/" + re.match(path_pattern, pkg_url).group(1).lower() + (".pro" if project else ".pri")

	return pkg_url, pkg_branch, pkg_path


def get_all_tags(pkg_url, branches=False, tags=True, allow_empty=False, allow_error=False):
	ls_args = ["git", "ls-remote", "--refs"]
	if not branches and tags:
		ls_args.append("--tags")
	elif not tags and branches:
		ls_args.append("--heads")
	if not allow_empty:
		ls_args.append("--exit-code")
	elif not tags and not branches:
		return []  # Nothing to check for
	ls_args.append(pkg_url)

	ls_res = sub_run(ls_args, check=not allow_error, stdout=subprocess.PIPE, encoding="UTF-8")
	ref_pattern = re.compile(r'[a-fA-F0-9]+\s+refs\/(?:tags|heads)\/([^\s]+)')
	tags = []
	for match in re.finditer(ref_pattern, ls_res.stdout):
		tags.append(match.group(1))
	return tags


def get_latest_tag(pkg_url, allow_empty=False, allow_error=False):
	all_tags = get_all_tags(pkg_url, allow_empty=allow_empty, allow_error=allow_error)
	if len(all_tags) > 0:
		return all_tags[-1]
	else:
		return None


def get_sources(pkg_url, pkg_branch, pull=True, clone=True):

	if pkg_branch is None:
		pkg_branch = get_latest_tag(pkg_url)

	cache_dir = get_cache_dir(pkg_url, pkg_branch)

	locker = LockFile(cache_dir)
	locker.acquire()
	try:
		needs_ro = False
		if path.isdir(path.join(cache_dir, ".git")):
			if pull and not path.exists(path.join(cache_dir, ".qdep_static_branch")):
				sub_run(["git", "pull", "--force", "--ff-only", "--update-shallow", "--recurse-submodules"], cwd=cache_dir, stdout=subprocess.DEVNULL, check=True)
				needs_ro = True
		elif clone:
			sub_run(["git", "clone", "--recurse-submodules", "--depth", "1", "--branch", pkg_branch, pkg_url, cache_dir], check=True)
			head_ref_res = sub_run(["git", "symbolic-ref", "HEAD"], cwd=cache_dir, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
			if head_ref_res.returncode != 0:
				open(path.join(cache_dir, ".qdep_static_branch"), 'a').close()
			needs_ro = True
		else:
			raise Exception("The --no-clone flag was specified - cannot install new packages with pulling disabled")

		if needs_ro:
			NO_WRITE_MASK = ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH
			for root, dirs, files in os.walk(cache_dir, topdown=True):
				dirs[:] = [d for d in dirs if d != ".git"]
				for file in files:
					f_path = path.join(root, file)
					cur_perm = stat.S_IMODE(os.lstat(f_path).st_mode)
					os.chmod(f_path, cur_perm & NO_WRITE_MASK)
	except:
		shutil.rmtree(cache_dir, ignore_errors=True)
		raise
	finally:
		locker.release()

	return cache_dir, pkg_branch


def extract_pro_depends(pro_file, qmake):
	packages = []

	with tempfile.TemporaryDirectory() as tmp_dir:
		dump_name = path.join(tmp_dir, "qdep_dummy.pro")
		with open(dump_name, "w") as dump_file:
			dump_file.write("depends += $$fromfile($$quote({}), QDEP_DEPENDS)\n".format(pro_file))
			dump_file.write("depends += $$fromfile($$quote({}), QDEP_PROJECT_SUBDIRS)\n".format(pro_file))
			dump_file.write("depends += $$fromfile($$quote({}), QDEP_PROJECT_LINK_DEPENDS)\n".format(pro_file))
			dump_file.write("depends += $$fromfile($$quote({}), QDEP_PROJECT_DEPENDS)\n".format(pro_file))
			dump_file.write("!write_file($$PWD/qdep_depends.txt, depends):error(\"write error\")\n")
		sub_run([qmake, dump_name], cwd=tmp_dir, check=True, stdout=subprocess.DEVNULL)
		with open(path.join(tmp_dir, "qdep_depends.txt"), "r") as dep_file:
			for line in dep_file.readlines():
				packages.append(line.strip())

	return packages


def eval_pro_depends(pro_file, qmake, make, dump_depends=False):
	with tempfile.TemporaryDirectory() as tmp_dir:
		print("Running {} on {}...".format(qmake, pro_file))
		sub_run([qmake] + (["CONFIG+=__qdep_dump_dependencies"] if dump_depends else []) + [pro_file], cwd=tmp_dir, check=True, stdout=subprocess.DEVNULL)
		print("Running {} qmake_all...".format(make))
		sub_run([make, "qmake_all"], cwd=tmp_dir, check=True, stdout=subprocess.DEVNULL)

		if dump_depends:
			all_deps = {}
			for root, dirs, files in os.walk(tmp_dir, topdown=True):
				for file in files:
					if file == "qdep_depends.txt":
						with open(path.join(root, file), "r") as dep_list:
							dep_name = dep_list.readline().strip()
							all_deps[dep_name] = []
							for dep in dep_list:
								all_deps[dep_name].append(dep.strip())

			return all_deps
		else:
			return None


def invert_map(original):
	inverted = OrderedDict()
	for key, values in original.items():
		for value in values:
			if value not in inverted:
				inverted[value] = [key]
			else:
				inverted[value].append(key)
	return inverted


def check_for_updates(packages):
	pkg_all = []
	pkg_new = {}
	for package in packages:
		pkg_url, pkg_version, _p = package_resolve(package)
		if pkg_version is None:
			pkg_all.append(package)
			continue

		# check if the package actually has any tags
		all_tags = get_all_tags(pkg_url, allow_empty=True)
		if len(all_tags) == 0:
			pkg_all.append(package)
			continue

		# check if actually a tag and not a branch
		if pkg_version not in all_tags and pkg_version in get_all_tags(pkg_url, branches=True, tags=False, allow_empty=True):
			pkg_all.append(package)
			continue

		# check if latest tag has changed
		if all_tags[-1] != pkg_version:
			pkg_name, _v, pkg_path = package_resolve(package, expand=False)
			print("Found a new version for package {}: {} -> {}".format(pkg_name, pkg_version, all_tags[-1]))
			new_pkg = "{}@{}{}".format(pkg_name, all_tags[-1], pkg_path)
			pkg_all.append(new_pkg)
			pkg_new[package] = new_pkg
		else:
			pkg_all.append(package)

	return pkg_all, pkg_new


def replace_or_print_update(pro_file, pkg_all, pkg_new, replace):
	if replace:
		with open(pro_file, "r") as in_file:
			with open(pro_file + ".qdepnew", "w") as out_file:
				for line in in_file:
					for src, repl in pkg_new.items():
						line = line.replace(src, repl)
					out_file.write(line)

		os.remove(pro_file)
		os.rename(pro_file + ".qdepnew", pro_file)
		print("Updated dependencies in {} - please check the file to make sure it has not been corrupted!".format(pro_file))
	else:
		print("")
		print("Dependencies for {}:".format(pro_file))
		print("QDEP_DEPENDS =", format(" \\\n\t".join(pkg_all)))



import setuptools
from qdep.qdep import version

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name="qdep",
	version=version,
	author="Skycoder42",
	author_email="skycoder42.de@gmx.de",
	description="A very basic yet simple to use dependency management tool for qmake based projects",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/Skycoder42/qdep",
	packages=setuptools.find_packages(),
	install_requires=[
		"appdirs",
		"lockfile",
		"argcomplete"
	],
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: BSD License",
		"Operating System :: OS Independent",
	],
	entry_points={
		"console_scripts": [
			'qdep=qdep.internal.cli:main'
		]
	}
)
